#!/usr/bin/env python3

"""
Base history mechanism
"""

import collections
import contextlib
import logging

from PyQt5 import QtCore


class HistoryVersion(collections.namedtuple('HistoryVersion', ['major', 'minor'])):
    """
    History version. Internal use only.
    """
    def __str__(self):
        return '.'.join(map(str, self))


class ConsistencyError(Exception):
    """
    Thrown when something bad happens (trying to undo with an empty
    past list for instance).
    """

class History(QtCore.QObject):
    """
    Maintains lists of past and future commands as they are done and
    undone. The `changed` signal is emitted whenever the history state
    changes. This class also maintains a pointer to a 'saved' state
    and can tell if its current state is the saved one or not (see
    is_modified()).
    """

    changed = QtCore.pyqtSignal()

    logger = logging.getLogger('pyqtcmd.History')

    def __init__(self):
        self.__past = []
        self.__future = []
        self.__version = HistoryVersion(0, 0)
        self.__saved_version = self.__version
        super().__init__()

    @contextlib.contextmanager
    def freeze(self): # pylint: disable=R0201
        """
        This context manager will wrap calls to
        Command.(undo|redo|do). You can override it to provide some
        sort of batch updating/signalling in your own project.
        """
        yield

    def check(self):
        """
        This just emits the changed signal; use it when you think
        UICommand states should be updated, when add_check_signal is
        not enough.
        """
        self.changed.emit()

    def reset(self, is_new=True):
        """
        Resets the history state. This is typically called when a new
        document has been loaded or created. If `is_new` is False, the
        saved state is not reset, so is_modified() will still return
        True. This is intended for session management, when loading a
        document that hasn't actually been saved yet.
        """
        self.__version = self.__version._replace(major=self.__version.major + 1)
        self.__past = []
        self.__future = []
        if is_new:
            self.__saved_version = self.__version
        self.logger.debug('Reset to version %s (saved %s)', self.__version, self.__saved_version)
        self.changed.emit()

    def is_modified(self):
        """
        Returns True if the history state is different from what it
        was when save_point() was last called (or when constructed, or
        reset with is_new=False).
        """
        return self.__version != self.__saved_version

    def save_point(self):
        """
        Sets the 'saved' state as the current one. Call this after
        saving a document to disk for instance.
        """
        self.__saved_version = self.__version
        self.logger.debug('Saved (version is now %s/%s)', self.__version, self.__saved_version)
        self.changed.emit()

    def run(self, command):
        """
        Run a command and put it in the past. In order to keep things
        consistent, the future list is deleted.
        """
        with self.freeze():
            self.logger.debug('Will run command "%s"', command)
            command.do()
            self.logger.debug('Ran command "%s"', command)
        self.__past.append(command)
        self.__future = []
        self.__version = self.__version._replace(minor=self.__version.minor + 1)
        self.logger.debug('Version is now %s/%s', self.__version, self.__saved_version)
        self.changed.emit()

    def can_undo(self):
        """Returns True if there is at least one past command"""
        return bool(self.__past)

    def undo_label(self):
        """The text for the current undo command, or None"""
        if self.__past:
            return self.__past[-1].label()
        return ''

    def can_redo(self):
        """Returns True if there is at least one future command"""
        return bool(self.__future)

    def redo_label(self):
        """The text for the current redo command, or None"""
        if self.__future:
            return self.__future[-1].label()
        return ''

    def undo(self):
        """
        Undo the latest command that was run. Throws ConsistencyError
        if there are no commands to undo.
        """
        if not self.__past:
            raise ConsistencyError('No command to undo')
        cmd = self.__past.pop()
        with self.freeze():
            self.logger.debug('Will undo "%s"', cmd)
            cmd.undo()
            self.logger.debug('Undid "%s"', cmd)
        self.__future.append(cmd)
        self.__version = self.__version._replace(minor=self.__version.minor - 1)
        self.logger.debug('Version is now %s/%s', self.__version, self.__saved_version)
        self.changed.emit()

    def redo(self):
        """
        Redo the latest command that was undone. Throws
        ConsistencyError if there are no command to redo.
        """
        if not self.__future:
            raise ConsistencyError('No command to redo')
        cmd = self.__future.pop()
        with self.freeze():
            self.logger.debug('Will redo "%s"', cmd)
            cmd.redo()
            self.logger.debug('Redid "%s"', cmd)
        self.__past.append(cmd)
        self.__version = self.__version._replace(minor=self.__version.minor + 1)
        self.logger.debug('Version is now %s/%s', self.__version, self.__saved_version)
        self.changed.emit()



#!/usr/bin/env python3

"""
Integration of commands with UI elements
"""

from PyQt5 import QtWidgets


class UICommand(QtWidgets.QAction):
    """
    An UICommand is the intermediate between UI elements and the
    underlying Commands. A Command encapsulates a single behavior; an
    UICommand has a longer life time since it's associated to a UI
    element. Its role is to instantiate and run the right Command when
    the user interacts with the element. It also has an
    enabled/disabled state that you can manage by overriding the
    should_be_enabled() method.
    """
    def __init__(self, parent=None, *, text=None, icon=None, shortcut=None, tip=None):
        super().__init__(parent)
        self.triggered.connect(self.do)
        self._check()
        self.history().changed.connect(self._check)

        if text is not None:
            self.setText(text) # pragma: no cover
        if icon is not None:
            self.setIcon(icon) # pragma: no cover
        if shortcut is not None:
            self.setShortcut(shortcut) # pragma: no cover
        if tip is not None:
            self.setToolTip(tip)
            self.setStatusTip(tip)

    def history(self):
        """Override this to return the History object"""
        raise NotImplementedError # pragma: no cover

    def add_signal_check(self, signal):
        """
        Call this, passing it a signal that could change the UIAction
        state (enabled/disabled, etc) when it's fired. The signal's
        signature does not matter.
        """
        signal.connect(self._check)

    def should_be_enabled(self): # pylint: disable=R0201
        """Return False from this method if the underlying UI element should be disabled."""
        return True

    def do(self):
        """
        Subclasses should override this to do whatever is necessary
        (for instance show a dialog for choosing a file name, etc) and
        then instantiate the right Command class and run() it.
        """
        raise NotImplementedError # pragma: no cover

    def set_button(self, btn):
        """
        Associates this command to a push button.
        """
        def changed():
            btn.setText(self.text())
            btn.setIcon(self.icon())
            btn.setEnabled(self.isEnabled())
            btn.setToolTip(self.toolTip())
            btn.setStatusTip(self.statusTip())
            btn.setVisible(self.isVisible())
        btn.clicked.connect(self.do)
        self.changed.connect(changed)
        changed()

    def _check(self, *unused_):
        self.setEnabled(self.should_be_enabled())


class UndoUICommandMixin:
    """
    Undo UICommand. This is a mixin so you can inherit from your concrete UICommand subclass.
    """
    def __init__(self, *args, text=None, **kwargs):
        self.__original_text = text
        super().__init__(*args, text=text, **kwargs)

    def do(self): # pylint: disable=C0111
        self.history().undo()

    def should_be_enabled(self): # pylint: disable=C0111
        return self.history().can_undo()

    def _check(self, *unused_):
        label = self.history().undo_label()
        if label is not None:
            self.setText(label)
        else:
            self.setText(self.__original_text)
        super()._check()


class RedoUICommandMixin:
    """
    Redo UICommand. This is a mixin so you can inherit from your concrete UICommand subclass.
    """
    def __init__(self, *args, text=None, **kwargs):
        self.__original_text = text
        super().__init__(*args, text=text, **kwargs)

    def do(self): # pylint: disable=C0111
        self.history().redo()

    def should_be_enabled(self): # pylint: disable=C0111
        return self.history().can_redo()

    def _check(self, *unused_):
        label = self.history().redo_label()
        if label is not None:
            self.setText(label)
        else:
            self.setText(self.__original_text)
        super()._check()


class NeedsSelectionUICommandMixin:
    """
    A mixin for UICommand. The constructor takes a `container` keyword
    argument. This must be an object which implements a
    `selectionChanged` signal (of any signature) and a `selection()`
    method, which should return the selection as an object with a
    __len__.
    """

    def __init__(self, parent=None, *, container, **kwargs):
        self._container = container
        super().__init__(parent, **kwargs)
        self.add_signal_check(container.selectionChanged)

    def container(self):
        """
        Accessor for the underlying container
        """
        return self._container

    def selection(self):
        """
        Returns the current selection. If you want to implement more
        drastic limitations than 'selection is not empty' you can
        override this to filter the actual underlying selection. The
        action will be enabled iff this returns a non-empty iterable.
        """
        return self.container().selection()

    def should_be_enabled(self): # pylint: disable=C0111
        return len(self.selection()) != 0 and super().should_be_enabled()



#!/usr/bin/env python3

"""
Base and generic commands
"""

class Command:
    """
    Command objects encapsulate a change to the document/project/etc's
    state. You should provide an __str__ method for logging purposes.
    """

    def do(self):
        """Override this to implement the change"""
        raise NotImplementedError # pragma: no cover

    def undo(self):
        """Override this to implement the change undo"""
        raise NotImplementedError # pragma: no cover

    def redo(self):
        """Override this to implement the change redo (calls do() by default)"""
        self.do()

    def label(self):
        """Provide a description of the command, to be used in menu items and tooltips"""


class UpdateStateCommand(Command):
    """
    A Command that updates an object's state through its
    __getstate__/__setstate__ special methods
    """

    def __init__(self, target, **state):
        super().__init__()
        self.__target = target
        self.__old = dict(target.__getstate__())
        self.__new = dict(self.__old)
        self.__new.update(state)

    def do(self):
        self.__target.__setstate__(self.__new)

    def undo(self):
        self.__target.__setstate__(self.__old)

    def __str__(self):
        return 'Update state of %s from %s to %s' % (self.__target, self.__old, self.__new)


class CompositeCommand(Command):
    """
    A Command that groups several other commands together.
    """

    def __init__(self):
        super().__init__()
        self.__commands = []

    def add_command(self, cmd):
        """
        Adds a new command to the chain. Commands will be run in
        order, and undone in reverse order.
        """
        self.__commands.append(cmd)

    def do(self):
        for cmd in self.__commands:
            cmd.do()

    def undo(self):
        for cmd in reversed(self.__commands):
            cmd.undo()

    def __str__(self):
        return 'Composition of: %s' % ', '.join(map(str, self.__commands))
